from __future__ import annotations

import os
import rich
import tvm
from rich.progress import Progress
from tvm import relay, IRModule, autotvm
from tvm.contrib.graph_executor import GraphModule
import onnx
import onnx_graphsurgeon as gs
from tvm import auto_scheduler
from pathlib import Path
from typing import TYPE_CHECKING, Union
from enum import Enum
from tvm import autotvm
if TYPE_CHECKING:
    Path = Union[Path, str]


class TunerStr(str, Enum):
    xgb = "xgb"
    random = "random"
    gridsearch = "gridsearch"


def str2tuner(tuner_str: TunerStr, task):
    if tuner_str == TunerStr.xgb:
        tuner = autotvm.tuner.XGBTuner(task, loss_type="rank")
    elif tuner_str == "ga":
        tuner = autotvm.tuner.GATuner(task, pop_size=50)
    elif tuner_str == "random":
        tuner = autotvm.tuner.RandomTuner(task)
    elif tuner_str == "gridsearch":
        tuner = autotvm.tuner.GridSearchTuner(task)
    else:
        raise ValueError(f"Invalid tuner: {tuner_str}")
    return tuner


def load_onnx(path: Path, batch_size: int = 1, dtype: str = "int8") -> tuple[IRModule, dict[str, tvm.nd.NDArray]]:
    rich.print(f"load model from [blue]{path}[/blue]")
    onnx_model = onnx.load(path)
    onnx.checker.check_model(onnx_model)
    graph = gs.import_onnx(onnx_model)
    shape_dict = {inp.name: list(map(lambda _: _ if _ > 0 else batch_size, inp.shape)) for inp in graph.inputs}
    mod, params = relay.frontend.from_onnx(onnx_model, shape=shape_dict, dtype=dtype)
    if not os.path.exists(".tvmer/ir"):
        os.makedirs(".tvmer/ir")
    with open(".tvmer/ir/origin", mode="w") as f:
        f.write(str(mod))
    return mod, params


def gen_library(mod, params, target: tvm.target.Target = "llvm", path: Path = ".tvmer/lib/compiled.so"):
    with tvm.transform.PassContext(opt_level=3):
        lib = relay.build(mod, target=target, params=params)
    if not os.path.exists(os.path.split(path)[0]):
        os.makedirs(os.path.split(path)[0])
    rich.print(f"export lib to [blue]{path}[/blue]")
    lib.export_library(path)
    return lib


def load_module(lib_path, dev):
    lib = tvm.runtime.load_module(lib_path)
    return GraphModule(lib['default'](dev))


def infer_time(lib_path: Path, input_data, dev=tvm.cpu(), repeat=10):
    import time
    module = load_module(lib_path, dev)

    t = time.time()
    for _ in range(repeat):
        module.run(**input_data)
    return (time.time() - t) / repeat


def tune(
        mod,
        params,
        target: tvm.target.Target = "llvm",
        num_measure_trials: int = 200
):
    tasks, task_weights = auto_scheduler.extract_tasks(mod["main"], params, target)

    for idx, task in enumerate(tasks):
        print("========== Task %d  (workload key: %s) ==========" % (idx, task.workload_key))
        print(task.compute_dag)

    tuner = auto_scheduler.TaskScheduler(tasks, task_weights)
    tune_option = auto_scheduler.TuningOptions(
        num_measure_trials=num_measure_trials,
        measure_callbacks=[auto_scheduler.RecordToFile(".tvmer/log/log_file")],
    )
    tuner.tune(tune_option)


def tune_with_template(
        mod,
        params,
        target: tvm.target.Target = "llvm",
        num_measure_trials: int = 200,
        tuner: TunerStr = TunerStr.xgb,
):
    tasks = autotvm.task.extract_from_program(
        mod["main"],
        target=target,
        params=params
    )

    log_filename = "tuning.log"
    tmp_log_file = "tuning.log.tmp"

    with Progress() as progress:
        task_tasks = progress.add_task("Task", total=len(tasks))
        for i, tsk in enumerate(reversed(tasks)):
            # create tuner
            tuner_obj = str2tuner(tuner, tsk)

            tsk_trial = min(num_measure_trials, len(tsk.config_space))

            prefix = "[Task %2d/%2d] " % (i + 1, len(tasks))
            task_tuner = progress.add_task("Tuning...", total=tsk_trial)

            # if use_transfer_learning:
            #     if os.path.isfile(tmp_log_file):
            #         tuner_obj.load_history(autotvm.record.load_from_file(tmp_log_file))
            # process tuning
            tuner_obj.tune(
                n_trial=tsk_trial,
                early_stopping=None,
                measure_option=autotvm.measure_option(
                    autotvm.LocalBuilder(),
                    autotvm.LocalRunner()
                ),
                callbacks=[
                    lambda cls, inputs, results: progress.update(task_tuner, advance=len(inputs)),
                    autotvm.callback.progress_bar(tsk_trial, prefix=prefix),
                    autotvm.callback.log_to_file(tmp_log_file),
                ],
            )
            progress.update(task_tasks, advance=1)

    # pick best records to a cache file
    autotvm.record.pick_best(tmp_log_file, log_filename)
    return autotvm.apply_history_best(log_filename)



from __future__ import annotations

import tvm
from tvm.contrib.graph_executor import GraphModule
import numpy as np
import typer
import rich
from rich.progress import Progress, ProgressBar
from pathlib import Path
from tvm import auto_scheduler
from tvm import autotvm

from tvmer.utils import load_onnx, gen_library, load_module, TunerStr
from tvmer import utils

app = typer.Typer(rich_markup_mode="rich")


@app.command()
def run(lib_path: Path, dev: str, repeat: int = 10):
    """
    run a compiled module
    """
    dev = tvm.device(dev)
    module = load_module(lib_path, dev)
    ftimer = module.module.time_evaluator("run", dev, number=10, repeat=repeat)
    prof_res = np.array(ftimer().results) * 1000
    rich.print(
        f"mean: {np.mean(prof_res):.2f} ms, std: {np.std(prof_res):.2f} ms"
    )
    # print(module.inp())
    # print(module.get_input_info())
    # module.set_input("479", np.zeros([1, 64, 8400]))
    # module.set_input("p1", np.zeros([1, 2, 8400]))


@app.command()
def compile(
        model_path: Path,
        target: str = "llvm",
        target_host: str = None,
        export_path: Path = ".tvmer/lib/default.so",
        dtype="float32"
):
    """
    compile a model
    """
    target = tvm.target.Target(target, host=target_host)

    mod, params = load_onnx(path=model_path, batch_size=1, dtype=dtype)
    gen_library(mod, params, target, export_path)


@app.command()
def tune(
        model_path: Path,
        target: str = "llvm",
        target_host: str = None,
        export_path: Path = ".tvmer/lib/tuned_default.so",
        dtype="float32",
        num_measure_trials: int = 200,
        tuner: TunerStr = None,
):
    """
    auto-tune a model (free template)
    """
    target = tvm.target.Target(target, host=target_host)

    print("Extract tasks...")
    mod, params = load_onnx(path=model_path, batch_size=1, dtype=dtype)
    if tuner is None:
        utils.tune(mod, params, target, num_measure_trials)
        gen_library(mod, params, target, export_path)
    else:
        with utils.tune_with_template(mod, params, target, num_measure_trials, tuner):
            gen_library(mod, params, target, export_path)


@app.command(deprecated=True)
def tune_next(
        model_path: Path,
        target: str = "llvm",
        target_host: str = None,
        export_path: Path = ".tvmer/lib/tuned_default.so",
        dtype="float32",
        num_measure_trials: int = 200,
):
    """
    auto-tune a model
    """
    rich.print("This command is deprecated.")
    print("Extract tasks...")
    target = tvm.target.Target(target, host=target_host)
    mod, params = load_onnx(path=model_path, batch_size=1, dtype=dtype)
    with utils.tune_with_template(mod, params, target, num_measure_trials):
        gen_library(mod, params, target, export_path)


# @app.command(deprecated=True)
# def main(target, dev, dtype="int8", lib_path: Path = "lib/arm_cpu_default.so"):
#     print("Extract tasks...")
#     mod, params = load_onnx(path="./model/yolov8s_detect.onnx", batch_size=1, dtype=dtype)
#     tune(mod, params, target)
#     lib = gen_library(mod, params, target, lib_path)
#
#     module = GraphModule(lib['default'](dev))
#     module.set_input("479", np.zeros([1, 64, 8400]))
#     ftimer = module.module.time_evaluator("run", dev, number=2, repeat=10)
#     prof_res = np.array(ftimer().results) * 1000  # multiply 1000 for converting to millisecond
#     print(
#         "%-20s %-19s (%s)" % ("rk3588", "%.2f ms" % np.mean(prof_res), "%.2f ms" % np.std(prof_res))
#     )
#
#     target = target if isinstance(target, str) else target.device_name
#     with open(f".tvmer/llvm_ir/{target}_{dtype}.source", mode="w") as f:
#         f.write(lib.get_lib().get_source())


if __name__ == '__main__':
    app()
